import json
import yaml
from cgi import parse_header
from io import BytesIO
from typing import Any, Callable, Dict, Optional, Tuple, Union

from .http_headers import HttpHeaders
from .http_message import (
    FormHttpMessage,
    HttpMessage,
    JsonHttpMessage,
    MultipartHttpMessage,
    SimpleHttpMessage,
    YamlHttpMessage,
)


class HttpParsedBodyTrait:
    _body: BytesIO
    _headers: HttpHeaders
    _parsed_body: Optional[Union[HttpMessage, Any]]
    _parsed_body_getter: Optional[Callable]
    _as_str: Optional[str]
    _as_dict: Optional[dict]

    @property
    def parsed_body(self) -> Union[HttpMessage, Any]:
        if self._parsed_body:
            return self._parsed_body

        if hasattr(self, "_parsed_body_getter"):
            self._parsed_body = self._parsed_body_getter()  # type: ignore
            return self._parsed_body

        content_type: Tuple[str, Dict[str, str]] = parse_header(self._headers["Content-Type"])  # type: ignore

        parsed_body: HttpMessage

        if content_type[0] == "multipart/form-data":
            parsed_body = MultipartHttpMessage.from_bytes(
                self._body,
                content_type[1].get("boundary", ""),
                content_type[1].get("charset", "utf8"),
            )
        elif content_type[0] == "application/x-www-form-urlencoded":
            parsed_body = FormHttpMessage.from_bytes(self._body, content_type[1].get("charset", "utf8"))

        elif content_type[0] == "application/json":
            parsed_body = JsonHttpMessage.from_bytes(self._body, content_type[1].get("charset", "utf8"))
        elif content_type[0] in (
            "text/vnd.yaml",
            "text/yaml",
            "text/x-yaml",
            "application/x-yaml",
        ):
            parsed_body = YamlHttpMessage.from_bytes(self._body, content_type[1].get("charset", "utf8"))
        else:
            self._body.seek(0)
            parsed_body = SimpleHttpMessage(self._body.read().decode(content_type[1].get("charset", "utf8")))

        self._parsed_body = parsed_body
        return self._parsed_body

    def as_str(self) -> str:
        if not self._as_str:
            self._body.seek(0)
            self._as_str = self._body.read().decode("utf8")

        return self._as_str

    def as_dict(self) -> dict:
        if self._as_dict is None:
            body_str = self.as_str()
            try:
                self._as_dict = json.loads(body_str)  # type: ignore

                return self._as_dict  # type: ignore
            except Exception:
                try:
                    self._as_dict = yaml.safe_load_all(body_str)  # type: ignore

                    return self._as_dict  # type: ignore
                except Exception:
                    self._as_dict = {}

        return self._as_dict



from enum import Enum
from typing import List, Dict


class SortDirection(Enum):
    ASCENDING = 1
    DESCENDING = -1

    def __str__(self) -> str:
        if self.value == -1:
            return "-"
        return "+"


def parse_sorting(value: str, allowed_fields: List[str] = None) -> Dict[str, SortDirection]:
    fields = value.split(",")
    result = {}

    for field in fields:
        field = field.strip()
        if field.startswith("-"):
            field_name = field[1:]
            direction = SortDirection.DESCENDING
        elif field.startswith("+"):
            field_name = field[1:]
            direction = SortDirection.ASCENDING
        else:
            field_name = field
            direction = SortDirection.ASCENDING

        if allowed_fields and field_name not in allowed_fields:
            continue

        result[field_name] = direction

    return result



from .http.http_error import HttpError, NotFoundError, BadRequestError


class ApplicationError(RuntimeError):
    @classmethod
    def for_invalid_namespace(cls, namespace: str) -> "ApplicationError":
        return ApplicationError(f"Failed to use namespace `{namespace}`")


__all__ = ["ApplicationError", "HttpError", "NotFoundError", "BadRequestError"]



from .query_criteria import QueryCriteria
from .expression import (
    Expression,
    EqualExpression,
    InExpression,
    LowerThanExpression,
    GreaterThanExpression,
    RangeExpression,
    PatternExpression,
)
from .sorting import SortDirection



from copy import copy
from io import BytesIO
from typing import Any, Dict, Optional, Union

from .http_body import write_body
from .http_cookies import HttpCookieJar, parse_cookie_header
from .http_headers import HttpHeaders
from .http_method import HttpMethod
from .http_parsed_body import HttpParsedBodyTrait
from .http_query_string import HttpQueryString
from chocs.routing import Route


class HttpRequest(HttpParsedBodyTrait):
    def __init__(
        self,
        method: Union[HttpMethod, str],
        path: str = "/",
        body: Union[BytesIO, bytes, bytearray, str, None] = None,
        query_string: Union[Optional[HttpQueryString], str] = None,
        headers: Union[Optional[HttpHeaders], Dict[str, str]] = None,
        encoding: str = "utf-8",
    ):
        if isinstance(method, str):
            method = HttpMethod(method.upper())

        if isinstance(query_string, str):
            query_string = HttpQueryString(query_string)

        if isinstance(headers, dict):
            headers = HttpHeaders(headers)

        self.method = method
        self.path = path
        self.query_string = query_string if query_string else HttpQueryString("")
        self.path_parameters: Dict[str, str] = {}
        self.route: Optional[Route] = None  # type: ignore
        self.attributes: Dict[str, Any] = {}
        self.encoding = encoding
        self._headers = headers if headers else HttpHeaders()
        self._cookies: Optional[HttpCookieJar] = None
        self._body = BytesIO()
        self._parsed_body = None
        self._as_dict = None
        self._as_str = None

        if body:
            write_body(self._body, body, encoding)

    @property
    def body(self) -> BytesIO:
        return copy(self._body)

    @property
    def cookies(self) -> HttpCookieJar:
        if self._cookies is None:
            self._cookies = parse_cookie_header(str(self._headers.get("cookie")))

        return copy(self._cookies)

    def __eq__(self, other) -> bool:
        if not isinstance(other, HttpRequest):
            return False

        return (
            self.method == other.method
            and self._headers == other._headers
            and self.path == other.path
            and self.query_string == other.query_string
            and self.body.getbuffer().nbytes == other.body.getbuffer().nbytes
        )

    def __str__(self) -> str:
        self._body.seek(0)
        return self._body.read().decode(self.encoding)

    @property
    def headers(self) -> HttpHeaders:
        return self._headers


__all__ = ["HttpRequest"]



from .aws import (
    AwsContext,
    AwsEvent,
    AwsServerlessFunction,
    create_http_request_from_aws_event,
)
from .serverless import IS_AWS_ENVIRONMENT, ServerlessFunction
from .wrapper import create_serverless_function



from io import BytesIO
from typing import Union


def write_body(
    body: BytesIO,
    contents: Union[str, bytes, bytearray, BytesIO],
    encoding: str = "utf8",
) -> None:
    if isinstance(contents, str):
        body.write(contents.encode(encoding))
    elif isinstance(contents, BytesIO):
        contents.seek(0)
        body.write(contents.read())
    else:
        body.write(contents)



from typing import Callable, Dict, Any

from gunicorn.app.base import BaseApplication


class GunicornApplication(BaseApplication):
    def __init__(self, wsgi_handler: Callable, options: Dict[str, Any] = None):
        self.options = options or {}
        self.wsgi_handler = wsgi_handler
        super().__init__()

    def load_config(self) -> None:
        config = {key: value for key, value in self.options.items() if key in self.cfg.settings and value is not None}

        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self) -> Any:
        return self.wsgi_handler

    def init(self, parser, opts, args):
        pass


def wsgi_serve(
    wsgi_handler: Callable,
    host: str = "127.0.0.1",
    port: int = 80,
    workers: int = 1,
) -> None:
    options = {
        "bind": f"{host}:{port}",
        "workers": workers,
    }
    server = GunicornApplication(wsgi_handler, options)
    server.run()



from typing import Callable

from bjoern import run


def wsgi_serve(
    wsgi_handler: Callable,
    host: str = "127.0.0.1",
    port: int = 80,
    workers: int = 1,
) -> None:
    run(wsgi_handler, host, port)



from abc import ABC, abstractmethod
from copy import deepcopy
from queue import Queue
from typing import Callable, Union

from chocs.http.http_request import HttpRequest
from chocs.http.http_response import HttpResponse


class MiddlewareHandler(ABC):
    @abstractmethod
    def __call__(self, request: HttpRequest) -> HttpResponse:
        ...


class Middleware(ABC):
    @abstractmethod
    def handle(self, request: HttpRequest, next: MiddlewareHandler) -> HttpResponse:
        ...


MiddlewareFunction = Callable[[HttpRequest, MiddlewareHandler], HttpResponse]


class MiddlewareCursor(MiddlewareHandler):
    def __init__(self, queue: Queue, parent: MiddlewareHandler):
        self.queue: Queue[Middleware] = Queue()
        self.queue.queue = deepcopy(queue.queue)
        self.parent = parent

    def __call__(self, request: HttpRequest) -> HttpResponse:
        if self.queue.empty():
            return self.parent(request)

        middleware: Union[Middleware, MiddlewareFunction] = self.queue.get()
        next = MiddlewareCursor(self.queue, self)

        if isinstance(middleware, Middleware):
            return middleware.handle(request, next)
        return middleware(request, next)


class EmptyPipelineHandler(MiddlewareHandler):
    def __call__(self, request: HttpRequest) -> HttpResponse:
        raise RuntimeError("Middleware pipe is empty.")


class MiddlewarePipeline(MiddlewareHandler, Middleware):
    def __init__(self, queue: Queue = Queue()):
        self.queue: Queue = Queue()
        self.queue.queue = deepcopy(queue.queue)

    def append(self, *middleware: Union[Middleware, Callable]) -> None:
        for item in middleware:
            self.queue.put(item)

    def __call__(self, request: HttpRequest) -> HttpResponse:
        return self.handle(request, EmptyPipelineHandler())

    def handle(self, request: HttpRequest, next: MiddlewareHandler) -> HttpResponse:
        return (MiddlewareCursor(self.queue, next)).__call__(request)

    @property
    def empty(self) -> bool:
        return self.queue.qsize() <= 0


__all__ = [
    "MiddlewareHandler",
    "Middleware",
    "MiddlewareFunction",
    "MiddlewareCursor",
    "MiddlewarePipeline",
]



from chocs.http.http_error import HttpError
from chocs.http.http_request import HttpRequest
from chocs.http.http_response import HttpResponse
from chocs.routing import Router
from chocs.serverless.serverless import ServerlessFunction

from .middleware import Middleware, MiddlewareHandler


class ApplicationMiddleware(Middleware):
    def __init__(self, router: Router):
        self.router = router

    def handle(self, request: HttpRequest, next: MiddlewareHandler) -> HttpResponse:
        try:
            handler = request.attributes["__handler__"]

            response: HttpResponse
            if isinstance(handler, ServerlessFunction):
                response = handler.function(request)
            else:
                response = handler(request)

            return response
        except HttpError as error:
            return HttpResponse(status=error.status_code, body=error.http_message)


__all__ = ["ApplicationMiddleware"]
